# -*- coding: utf-8 -*-
# pylint: disable=invalid-name
"""
 Used to proxy Aliyun DLA Spark API to standard Livy 0.7.0 API
"""
import argparse
import json
import logging
import time
from argparse import Namespace
from datetime import date

from aliyunsdkcore.acs_exception.exceptions import ServerException
from aliyunsdkcore.client import AcsClient
from aliyunsdkopenanalytics_open.request.v20180619 import \
    KillSparkJobRequest, \
    ListSparkStatementsRequest, \
    ExecuteSparkStatementRequest, \
    GetSparkStatementRequest, \
    GetJobLogRequest, \
    CancelSparkStatementRequest, \
    GetSparkSessionStateRequest, \
    GetJobDetailRequest, \
    ListSparkJobRequest, \
    SubmitSparkJobRequest

from gevent.pywsgi import WSGIServer
from flask import Flask, request
from flask import jsonify

# flask app
from proxy.utils import job_id_to_session_id, session_id_to_job_id

app = Flask(__name__)
# current software version
_VERSION = 'Aliyun DLA Spark Proxy 1.0.0'
# create logger
logger = logging.getLogger('DLA')
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s-%(name)s-%(levelname)s: %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


@app.route('/')
def index():
    """
    :return: Current Version
    """
    resp = {
        "version": _VERSION
    }
    return jsonify(resp)

def get_job_log(job_id: str):
    """
    Return the log of this dla spark job
    :param job_id: dla job id
    :return: job status
    """
    req = GetJobLogRequest.GetJobLogRequest()
    req.set_JobId(job_id)
    req.set_VcName(vc_name)
    origin_log: list = []
    try:
        result = json.loads(ali_client.do_action_with_exception(req))
        logger.debug(result)
        origin_log = str(result['Data']).split('\n')
    except ServerException as e:
        logger.debug(e.message)
    return origin_log

def get_job_status(job_id: str):
    """
    Return the status of this dla spark job

    :param job_id: dla job id
    :return: job log
    """
    # get detail
    req = GetJobDetailRequest.GetJobDetailRequest()
    req.set_JobId(job_id)
    req.set_VcName(vc_name)

    try:
        job_detail = json.loads(ali_client.do_action_with_exception(req))
        logger.debug(job_detail)
    except ServerException as se:
        logger.error("%s, %s, %s", se.message, se.request_id, se.error_code)
        raise
    logger.debug(job_detail)
    job_detail = job_detail['JobDetail']

    # get state
    req = GetSparkSessionStateRequest.GetSparkSessionStateRequest()
    req.set_JobId(job_id)
    try:
        session_state = json.loads(ali_client.do_action_with_exception(req))
        logger.debug(session_state)
        _current_state = session_state['State']
    except ServerException as se:
        logger.warning("%s, %s, %s", se.message, se.request_id, se.error_code)
        # service not ready means that job is starting, it will go to running or dead
        if se.error_code == "ServiceNotReady":
            _current_state = "starting"
        # empty result means that job return empty string, wait it go to normal response or dead
        elif se.error_code == "EmptyResult":
            _current_state = "starting"
        else:
            # start more than n min, default 5 min, can set by --timeout
            if int(time.time() * 1000) - job_detail['CreateTimeValue'] > job_ttl * 1000:
                raise
            else:
                _current_state = "starting"

    # update local info
    live_sessions[job_id] = _current_state
    return _current_state, job_detail


@app.route('/sessions', methods=['GET'])
def get_sessions():
    """
    return active sessions
    :rtype: object
    """
    page_size = request.args.get('size', 100)
    page_num = int(request.args.get('from', 1)) / int(page_size) + 1
    session_list = []
    total_num = 0
    try:
        # local cache starting sessions
        for job_id in live_sessions:
            if live_sessions[job_id] == 'starting':
                job_status, job_detail = get_job_status(job_id)

                # status change to running or dead
                if job_status != 'starting':
                    continue

                # append to final result
                session = {
                    "id": job_id_to_session_id(job_id),
                    "name": job_detail['JobName'],
                    "appId": job_id,
                    "owner": None,
                    "proxyUser": None,
                    "state": job_status,
                    "kind": "python",
                    "appInfo": {
                        "driverLogUrl": None,
                        "sparkUiUrl": job_detail['SparkUI']
                    },
                    "log": get_job_log(job_id)
                }
                session_list.append(session)
                total_num += 1

        # get running sessions
        req = ListSparkJobRequest.ListSparkJobRequest()
        req.set_VcName(vc_name)
        req.set_PageNumber(int(page_num))
        req.set_PageSize(page_size)
        req.set_Condition('{"status": "running", "jobType": "session"}')
        result = json.loads(ali_client.do_action_with_exception(req))
        logger.debug(result)
        total_num += result['DataResult']['TotalCount']
        for item in result['DataResult']['JobList']:
            job_id = item['JobId']
            job_status, job_detail = get_job_status(item['JobId'])
            session = {
                "id": job_id_to_session_id(job_id),
                "name": job_detail['JobName'],
                "appId": job_id,
                "owner": None,
                "proxyUser": None,
                "state": job_status,
                "kind": "python",
                "appInfo": {
                    "driverLogUrl": None,
                    "sparkUiUrl": job_detail['SparkUI']
                },
                "log": get_job_log(job_id)
            }
            session_list.append(session)
    except ServerException as se:
        logger.error("%s, %s, %s", se.message, se.request_id, se.error_code)
        return se.message, 400
    # some job may duplicate in this final result because it change from starting to running
    dedup_cache = []
    session_list_dedup = []
    for session in session_list:
        if session['appId'] in dedup_cache:
            continue
        else:
            session_list_dedup.append(session)
            dedup_cache.append(session['appId'])
    # return the result
    return json.dumps(
        {
            "from": request.args.get('from', 1),
            "total": total_num,
            "sessions": session_list_dedup[0: int(page_size)]
        }
    )


@app.route('/sessions', methods=['POST'])
def post_sessions():
    """
    Create session
    :rtype: object
    """
    spark_conf: dict = request.get_json(force=True)
    logger.debug(spark_conf)
    # set job name
    if 'name' in spark_conf:
        job_name = spark_conf['name']
    else:
        job_name = "notebook_" + date.today().strftime("%Y-%m-%d")

    # set default start kind of session
    if "kind" in spark_conf:
        default_kind = spark_conf['kind']
    else:
        default_kind = 'python'

    dla_job_config = {
        "type": "session",
        "name": job_name,
    }

    if 'jars' in spark_conf:
        dla_job_config['jars'] = spark_conf['jars']

    if 'pyFiles' in spark_conf:
        dla_job_config['pyFiles'] = spark_conf['pyFiles']

    if 'conf' in spark_conf:
        dla_job_config['conf'] = spark_conf['conf']
    else:
        dla_job_config['conf'] = {}

    if 'files' in spark_conf:
        dla_job_config['files'] = spark_conf['files']

    if 'archives' in spark_conf:
        dla_job_config['archives'] = spark_conf['archives']

    dla_job_config['conf']['spark.dla.session.kind'] = default_kind
    logger.info(dla_job_config)
    req = SubmitSparkJobRequest.SubmitSparkJobRequest()
    req.set_VcName(vc_name)
    req.set_ConfigJson(json.dumps(dla_job_config))
    try:
        result = json.loads(ali_client.do_action_with_exception(req))
    except ServerException as se:
        logger.debug(se)
        return se.message, 400
    job_id = result['JobId']
    session_id: int = int(job_id_to_session_id(job_id))
    resp = {
        "id": session_id,
        "name": job_name,
        "appId": result['JobId'],
        "owner": None,
        "proxyUser": None,
        "state": "starting",
        "kind": "python",
        "appInfo": {
            "driverLogUrl": None,
            "sparkUiUrl": None
        },
        "log": None
    }

    # update local cache
    live_sessions[job_id] = "starting"
    return json.dumps(resp), 201, {'content-type': 'application/json'}


@app.route('/sessions/<session_id>', methods=['GET'])
def get_session(session_id):
    """
    Get the running sessions from Aliyun
    :rtype: object
    """
    job_id: str = session_id_to_job_id(session_id)

    try:
        job_status, job_detail = get_job_status(job_id)
    except ServerException as se:
        logger.error("%s, %s, %s", se.message, se.request_id, se.error_code)
        return se.message, 400

    resp = {
        "id": session_id,
        "name": job_detail['JobName'],
        "appId": job_id,
        "owner": None,
        "proxyUser": None,
        "state": job_status,
        "kind": "python",
        "appInfo": {
            "driverLogUrl": None,
            "sparkUiUrl": job_detail['SparkUI']
        },
        "log": get_job_log(job_id)
    }
    # return result
    return jsonify(resp)


@app.route('/sessions/<session_id>/state', methods=['GET'])
def get_session_state(session_id):
    """
    return basic
    :rtype: basestring
    """
    job_id: str = session_id_to_job_id(session_id)

    try:
        status, job_detail = get_job_status(job_id)
    except ServerException as se:
        logger.error("%s, %s, %s", se.message, se.request_id, se.error_code)
        return se.message, 400

    resp = {
        "id": int(session_id),
        "state": status
    }
    return jsonify(resp)


@app.route('/sessions/<session_id>', methods=['DELETE'])
def delete_session(session_id):
    """
    As livy, always kill session process
    :rtype: object
    """
    req = KillSparkJobRequest.KillSparkJobRequest()
    job_id: str = session_id_to_job_id(session_id)
    req.set_JobId(job_id)
    req.set_VcName(vc_name)
    try:
        json.loads(ali_client.do_action_with_exception(req))
    except ServerException as se:
        logger.error("%s, %s, %s", se.message, se.request_id, se.error_code)
    except Exception as e:
        logger.error(str(e))
    live_sessions[job_id] = "killed"
    return "killed"


@app.route('/sessions/<session_id>/log', methods=['GET'])
def get_session_log(session_id):
    """
    Get spark driver log
    :rtype: object
    """
    job_id: str = session_id_to_job_id(session_id)
    origin_log = get_job_log(job_id)
    # always return tail log
    resp = {
        "id": int(session_id),
        "from": 0,
        "size": len(origin_log),
        "log": origin_log
    }
    return jsonify(resp)


@app.route('/sessions/<session_id>/statements', methods=['GET'])
def get_session_statements(session_id):
    """
    Get all statements from a running session
    :rtype: object
    """
    req = ListSparkStatementsRequest.ListSparkStatementsRequest()
    job_id: str = session_id_to_job_id(session_id)
    req.set_JobId(job_id)
    try:
        result = json.loads(ali_client.do_action_with_exception(req))
        logger.debug(result)
    except ServerException as se:
        logger.error("%s, %s, %s", se.message, se.request_id, se.error_code)
        return se.message, 400

    resp = []
    for stmt in result['Statements']:
        resp.append(
            {
                "id": stmt['Id'],
                "code": stmt['Code'],
                'state': stmt['State'],
                'output': stmt['Output'],
                'progress': stmt['Progress'],
                "started": stmt['StartedTime'],
                "completed": stmt['CompletedTime']
            }
        )
    return jsonify(resp)


@app.route('/sessions/<session_id>/statements', methods=['POST'])
def post_session_statement(session_id):
    """
    Run code in session
    :rtype: object
    """
    code_form: dict = request.get_json(force=True)
    logger.debug(code_form)
    req = ExecuteSparkStatementRequest.ExecuteSparkStatementRequest()
    job_id: str = session_id_to_job_id(session_id)
    req.set_JobId(job_id)
    req.set_Code(code_form['code'])
    # same as livy, python is the default code kind
    if 'kind' in code_form:
        req.set_Kind(code_form['kind'])
    else:
        req.set_Kind('default')
    try:
        result = json.loads(ali_client.do_action_with_exception(req))
        logger.debug(result)
    except ServerException as e:
        logger.error(e.message)
        return e.message, 400
    # return created info
    stmtId = result['Data']
    resp = {
        "id": stmtId,
        "code": code_form['code'],
        'state': 'waiting',
        'output': None,
        'progress': 0,
        "started": 0,
        "completed": 0
    }
    return json.dumps(resp), 201, {'content-type': 'application/json'}


@app.route('/sessions/<session_id>/statements/<statement_id>', methods=['GET'])
def get_session_statement(session_id, statement_id):
    """
    get  statement by id
    :rtype: object
    """
    req = GetSparkStatementRequest.GetSparkStatementRequest()
    job_id: str = session_id_to_job_id(session_id)
    req.set_JobId(job_id)
    req.set_StatementId(int(statement_id))
    try:
        result = json.loads(ali_client.do_action_with_exception(req))
        logger.debug(result)
    except ServerException as e:
        logger.error(e.message)
        return e.message, 400

    stmt = result['Data']
    logger.debug(result)

    resp = {
        "id": stmt['Id'],
        "code": stmt['Code'],
        'state': stmt['State'],
        'progress': stmt.get('Process', 0),
        "started": stmt['StartedTime'],
        "completed": stmt['CompletedTime']
    }

    if 'Output' in stmt:
        resp['output'] = json.loads(stmt['Output'])
    return jsonify(resp)


@app.route('/sessions/<session_id>/statements/<statement_id>/cancel', methods=['POST'])
def cancel_session_statements(session_id, statement_id):
    """
    Same as livy, always return canceled
    :rtype: object
    """
    req = CancelSparkStatementRequest.CancelSparkStatementRequest()
    job_id: str = session_id_to_job_id(session_id)
    req.set_JobId(job_id)
    req.set_StatementId(int(statement_id))
    json.loads(ali_client.do_action_with_exception(req))

    resp = {
        "msg": "canceled"
    }
    return jsonify(resp)


def check_args():
    """
    check the args valid
    TODO: need to valid the vc access permission
    :rtype: object
    """
    logger.info("Aliyun DLA Proxy is ready")
    return


def main():
    """
    Server init function
    :rtype: object
    """
    parser = argparse.ArgumentParser(description='Proxy AliYun DLA as Livy')
    parser.add_argument('--vcname', type=str, required=True, help='Virtual Cluster Name')
    parser.add_argument('-i', '--access-key-id', dest='ak', type=str, required=True, help='Aliyun Access Key Id')
    parser.add_argument('-k', '--access-key-secret', dest='secret', type=str, required=True,
                        help='Aliyun Access Key Secret')
    parser.add_argument('--region', type=str, required=True, help='Aliyun Region Id')
    parser.add_argument('--host', default="127.0.0.1", type=str, help='Proxy Host Ip')
    parser.add_argument('--port', default=5000, type=int, help='Proxy Host Port')
    parser.add_argument('--loglevel', default="INFO", type=str, help='python standard log level')
    parser.add_argument('--timeout', default="300", type=str, help='time out for session job in seconds')
    args: Namespace = parser.parse_args()
    # set the global vc name
    global vc_name
    vc_name = args.vcname
    # init global client
    global ali_client
    ali_client = AcsClient(ak=args.ak, secret=args.secret, region_id=args.region)
    # live sessions  list
    global live_sessions
    live_sessions = {}
    # live sessions  list
    global job_ttl
    job_ttl = int(args.timeout)
    # validate the args
    check_args()
    logger.setLevel(args.loglevel)
    # run flask app
    http_server = WSGIServer((args.host, args.port), app)
    http_server.serve_forever()


if __name__ == '__main__':
    main()
